using System;
using System.Net;
using NUnit.Framework;

using ImapPseudoServer = Smdn.Net.Imap4.Tests.ImapPseudoServer;

namespace Smdn.Net.Imap4.WebClients {
  [TestFixture]
  public class ImapSearchMessageWebRequestTests {
    [SetUp]
    public void Setup()
    {
      ImapSessionManager.RegisterWebRequestPrefix();
    }

    private void Request(ImapPseudoServer server, string method, bool searchresCapable, string query, string[] methodResponses, Action<ImapWebRequest, ImapWebResponse> responseAction)
    {
      Request(server,
              method,
              searchresCapable
                ? new[] {"IMAP4REV1", "SEARCHRES"}
                : new[] {"IMAP4REV1"},
              query,
              methodResponses,
              null,
              responseAction);
    }

    private void Request(ImapPseudoServer server, string method, string[] capabilities, string query, string[] methodResponses, Action<ImapWebRequest, ImapWebResponse> responseAction)
    {
      Request(server,
              method,
              capabilities,
              query,
              methodResponses,
              null,
              responseAction);
    }

    private void Request(ImapPseudoServer server, string method, string[] capabilities, string query, string[] methodResponses, Action<ImapWebRequest> presetRequest, Action<ImapWebRequest, ImapWebResponse> responseAction)
    {
      server.Start();

      var request = WebRequest.Create(string.Format("imap://user@{0}/INBOX/{1}", server.HostPort, query)) as ImapWebRequest;

      Assert.AreEqual("ImapSearchMessageWebRequest", request.GetType().Name);

      request.Credentials = new NetworkCredential("user", "pass");
      request.KeepAlive = false;
      request.Timeout = 1000;
      request.Method = method;

      // greeting
      server.EnqueueResponse("* OK ready\r\n");
      // CAPABILITY
      server.EnqueueResponse(string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capabilities)) +
                             "0000 OK done\r\n");
      // LOGIN
      server.EnqueueResponse("0001 OK done\r\n");
      // SELECT
      server.EnqueueResponse("0002 OK done\r\n");

      int commandTag = 3;

      foreach (var methodResponse in methodResponses) {
        server.EnqueueResponse(methodResponse);

        commandTag++;
      }

      // CLOSE
      server.EnqueueResponse(string.Format("{0:x4} OK done\r\n", commandTag++));
      // LOGOUT
      server.EnqueueResponse("* BYE logging out\r\n" + 
                             string.Format("{0:x4} OK done\r\n", commandTag++));

      if (presetRequest != null)
        presetRequest(request);

      try {
        using (var response = request.GetResponse() as ImapWebResponse) {
          responseAction(request, response);
        }
      }
      catch (WebException ex) {
        if (ex.Status == WebExceptionStatus.ProtocolError)
          responseAction(request, ex.Response as ImapWebResponse);
        else
          throw ex;
      }

      server.DequeueRequest(); // CAPABILITY
      server.DequeueRequest(); // LOGIN
      server.DequeueRequest(); // SELECT
    }

    [Test]
    public void TestMethodNoOp()
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var request = WebRequest.Create(string.Format("imap://user@{0}/INBOX/?UID 1", server.HostPort)) as ImapWebRequest;

        Assert.AreEqual("ImapSearchMessageWebRequest", request.GetType().Name);

        request.Credentials = new NetworkCredential("user", "pass");
        request.KeepAlive = false;
        request.Timeout = 1000;
        request.Method = "NOOP";

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0000 OK done\r\n");
        // LOGIN
        server.EnqueueResponse("0001 OK done\r\n");
        // NOOP
        server.EnqueueResponse("0002 OK done\r\n");
        // CLOSE
        server.EnqueueResponse("0003 OK done\r\n");
        // LOGOUT
        server.EnqueueResponse("* BYE logging out\r\n" + 
                               "0004 OK done\r\n");

        using (var response = request.GetResponse() as ImapWebResponse) {
          Assert.IsNotNull(response);

          Assert.AreEqual(response.ResponseUri, request.RequestUri);

          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Succeeded);
        }

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // LOGIN
        StringAssert.Contains("NOOP", server.DequeueRequest());
      }
    }

    [Test]
    public void TestGetSearchResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Search);
    }

    [Test]
    public void TestGetSortResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Sort);
    }

    [Test]
    public void TestGetThreadResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Thread);
    }

    [Test]
    public void TestGetCopyResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Copy);
    }

    [Test]
    public void TestGetStoreResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Store);
    }

    [Test]
    public void TestGetExpungeResponseSelectFailure()
    {
      GetResponseSelectFailure(ImapWebRequestMethods.Expunge);
    }

    private void GetResponseSelectFailure(string method)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start();

        var request = WebRequest.Create(string.Format("imap://{0}/INBOX/?UID 1", server.HostPort)) as ImapWebRequest;

        request.KeepAlive = false;
        request.Timeout = 1000;
        request.Method = method;

        // greeting
        server.EnqueueResponse("* OK ready\r\n");
        // CAPABILITY
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0000 OK done\r\n");
        // LOGIN
        server.EnqueueResponse("0001 OK done\r\n");
        // SELECT
        server.EnqueueResponse("0002 NO failed\r\n");
        // LOGOUT
        server.EnqueueResponse("* BYE logging out\r\n" + 
                               "0003 OK done\r\n");

        try {
          using (var response = request.GetResponse()) {
          }
        }
        catch (WebException ex) {
          if (ex.Status != WebExceptionStatus.ProtocolError)
            throw ex;

          var resp = ex.Response as ImapWebResponse;

          Assert.AreEqual(request.RequestUri, resp.ResponseUri);

          Assert.IsNotNull(resp);
          Assert.IsTrue(resp.Result.Failed);
        }

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // LOGIN

        StringAssert.Contains("SELECT \"INBOX\"", server.DequeueRequest());
      }
    }

    [Test]
    public void TestGetSearchResponseUidSearchFailure()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Search, false);
    }

    [Test]
    public void TestGetSearchResponseUidSearchFailureSearchresCapable()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Search, true);
    }

    [Test]
    public void TestGetCopyResponseUidSearchFailure()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Copy, false);
    }

    [Test]
    public void TestGetCopyResponseUidSearchFailureSearchresCapable()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Copy, true);
    }

    [Test]
    public void TestGetStoreResponseUidSearchFailure()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Store, false);
    }

    [Test]
    public void TestGetStoreResponseUidSearchFailureSearchresCapable()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Store, true);
    }

    [Test]
    public void TestGetExpungeResponseUidSearchFailure()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Expunge, false);
    }

    [Test]
    public void TestGetExpungeResponseUidSearchFailureSearchresCapable()
    {
      GetResponseUidSearchFailure(ImapWebRequestMethods.Expunge, true);
    }

    private void GetResponseUidSearchFailure(string method, bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "0003 NO failed\r\n",
        };

        Request(server, method, searchresCapable, "?UID 1", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Failed);
        });

        if (searchresCapable)
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UID 1\r\n", server.DequeueRequest());
        else
          StringAssert.EndsWith("UID SEARCH UID 1\r\n", server.DequeueRequest());
      }
    }

#if false
    [Test]
    public void TestGetThreadResponse()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new [] {
          "* THREAD (2)(3 6 (4 23)(44 7 96))\r\n" +
          "0002 OK done\r\n",

          "0003 NO failed\r\n",
        };

        Request(server, ImapWebRequestMethods.Thread, "?SINCE 5-MAR-2000", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.IsNotNull(response);

          Assert.AreEqual(response.ResponseUri, request.RequestUri);

          Assert.IsNotNull(response.Result);
          //Assert.IsTrue(response.Result.Succeeded);
          Assert.IsTrue(response.Result.Failed);
        });

        StringAssert.Contains("THREAD REFERENCES UTF-8 SINCE 5-MAR-2000", server.DequeueRequest());
      }
    }
#endif
    [Test, Ignore("TODO: THREAD=ORDEREDSUBJECT incapable")]
    public void TestGetThreadResponseThreadOrderedSubjectIncapable()
    {
    }

    [Test, Ignore("TODO: THREAD=REFERENCES incapable")]
    public void TestGetThreadResponseThreadReferencesIncapable()
    {
    }

    [Test, Ignore("TODO: THREAD=REFS incapable")]
    public void TestGetThreadResponseThreadRefsIncapable()
    {
    }

    [Test, Ignore("TODO: SORT incapable")]
    public void TestGetSortResponseSortIncapable()
    {
    }

    [Test]
    public void TestGetSearchResponse()
    {
      GetSearchResponse(false);
    }

    [Test]
    public void TestGetSearchResponseSearchresCapable()
    {
      GetSearchResponse(true);
    }

    private void GetSearchResponse(bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          searchresCapable
            ? "0003 OK done\r\n"
            : "* SEARCH 3 4 5\r\n" + 
              "0003 OK done\r\n",
          // UID FETCH
          "* 1 FETCH (UID 3)\r\n" +
          "* 2 FETCH (UID 4)\r\n" +
          "* 3 FETCH (UID 5)\r\n" +
          "0004 OK done\r\n"
        };

        Request(server, ImapWebRequestMethods.Search, searchresCapable, "?UNSEEN", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Succeeded);

          Assert.IsNotNull(response.Messages);
          Assert.AreEqual(3, response.Messages.Length);

          var mailbox = new Uri(request.RequestUri.GetLeftPart(UriPartial.Path));

          Assert.AreEqual(new Uri(mailbox, ";UID=3"), response.Messages[0].Url);
          Assert.AreEqual(new Uri(mailbox, ";UID=4"), response.Messages[1].Url);
          Assert.AreEqual(new Uri(mailbox, ";UID=5"), response.Messages[2].Url);
        });

        if (searchresCapable) {
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID FETCH $ ALL\r\n", server.DequeueRequest());
        }
        else {
          StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID FETCH 3,4,5 ALL\r\n", server.DequeueRequest());
        }
      }
    }

    [Test]
    public void TestGetSearchResponseFailure()
    {
      GetSearchResponseFailure(false);
    }

    [Test]
    public void TestGetSearchResponseFailureSearchresCapable()
    {
      GetSearchResponseFailure(true);
    }

    private void GetSearchResponseFailure(bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          searchresCapable
            ? "0003 OK done\r\n"
            : "* SEARCH 3 4 5\r\n" + 
              "0003 OK done\r\n",
          // UID FETCH
          "0004 NO failed\r\n"
        };

        Request(server, ImapWebRequestMethods.Search, searchresCapable, "?UNSEEN", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Failed);

          Assert.IsNotNull(response.Messages);
          Assert.IsEmpty(response.Messages);
        });

        if (searchresCapable) {
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID FETCH $ ALL\r\n", server.DequeueRequest());
        }
        else {
          StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID FETCH 3,4,5 ALL\r\n", server.DequeueRequest());
        }
      }
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestGetCopyResponseDestinationUriNotSet()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
        };

        Request(server, ImapWebRequestMethods.Copy, false, "?UNSEEN", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.Fail("InvalidOperationException not thrown");
        });
      }
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestGetCopyResponseInvalidDestinationUriForm1()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
        };

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.DestinationUri = new Uri(new Uri(req.RequestUri.GetLeftPart(UriPartial.Authority)), "copyto-mailbox/;UID=1");
        };

        Request(server, ImapWebRequestMethods.Copy, new[] {"IMAP4REV1"}, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.Fail("InvalidOperationException not thrown");
        });
      }
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestGetCopyResponseInvalidDestinationUriForm2()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
        };

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.DestinationUri = new Uri(req.RequestUri.GetLeftPart(UriPartial.Authority));
        };

        Request(server, ImapWebRequestMethods.Copy, new[] {"IMAP4REV1"}, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.Fail("InvalidOperationException not thrown");
        });
      }
    }

    [Test]
    public void TestGetCopyResponseFailure()
    {
      GetCopyResponseFailure(false);
    }

    [Test]
    public void TestGetCopyResponseFailureSearchresCapable()
    {
      GetCopyResponseFailure(true);
    }

    private void GetCopyResponseFailure(bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          searchresCapable
            ? "0003 OK done\r\n"
            : "* SEARCH 3 4 5\r\n" + 
              "0003 OK done\r\n",
          // UID COPY
          "0004 NO [TRYCREATE] failed\r\n",
        };

        var capabilities = searchresCapable
          ? new[] {"IMAP4REV1", "SEARCHRES"}
          : new[] {"IMAP4REV1"};

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.AllowCreateMailbox = false;
          req.DestinationUri = new Uri(new Uri(req.RequestUri.GetLeftPart(UriPartial.Path)), "./コピー先メールボックス");
        };

        Request(server, ImapWebRequestMethods.Copy, capabilities, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Failed);
        });

        if (searchresCapable) {
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID COPY $ \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        }
        else {
          StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID COPY 3,4,5 \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        }
      }
    }

    [Test]
    public void TestGetCopyResponse()
    {
      GetCopyResponse(false);
    }

    [Test]
    public void TestGetCopyResponseSearchresCapable()
    {
      GetCopyResponse(true);
    }

    private void GetCopyResponse(bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          searchresCapable
            ? "0003 OK done\r\n"
            : "* SEARCH 3 4 5\r\n" + 
              "0003 OK done\r\n",
          // UID COPY
          "0004 OK done\r\n"
        };

        var capabilities = searchresCapable
          ? new[] {"IMAP4REV1", "SEARCHRES"}
          : new[] {"IMAP4REV1"};

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.AllowCreateMailbox = false;
          req.DestinationUri = new Uri(new Uri(req.RequestUri.GetLeftPart(UriPartial.Path)), "./コピー先メールボックス");
        };

        Request(server, ImapWebRequestMethods.Copy, capabilities, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Succeeded);

          Assert.IsNotNull(response.Mailboxes);
          Assert.AreEqual(0, response.Mailboxes.Length);
        });

        if (searchresCapable) {
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID COPY $ \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        }
        else {
          StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID COPY 3,4,5 \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        }
      }
    }

    [Test]
    public void TestGetCopyResponseAllowCreateMailbox()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
          // UID COPY
          "0004 NO [TRYCREATE] failed\r\n",
          // CREATE
          "0005 OK done\r\n",
          // UID COPY
          "0006 OK done\r\n",
          // SUBSCRIBE
          "0007 OK done\r\n",
          // LSUB
          "* LSUB () \".\" \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n" + 
          "0008 OK done\r\n"
        };

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.AllowCreateMailbox = true;
          req.DestinationUri = new Uri(new Uri(req.RequestUri.GetLeftPart(UriPartial.Path)), "./コピー先メールボックス");
        };

        Request(server, ImapWebRequestMethods.Copy, new[] {"IMAP4REV1"}, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Succeeded);

          Assert.IsNotNull(response.Mailboxes);
          Assert.AreEqual(1, response.Mailboxes.Length);
          Assert.AreEqual(new Uri(new Uri(request.RequestUri.GetLeftPart(UriPartial.Authority)), "INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-/"),
                          response.Mailboxes[0].Url);
        });

        StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
        StringAssert.EndsWith("UID COPY 3,4,5 \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        StringAssert.EndsWith("CREATE \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        StringAssert.EndsWith("UID COPY 3,4,5 \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        StringAssert.EndsWith("SUBSCRIBE \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        StringAssert.EndsWith("LSUB \"\" \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
      }
    }

    [Test]
    public void TestGetCopyResponseCreateDestinationMailboxFailure()
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
          // UID COPY
          "0004 NO [TRYCREATE] failed\r\n",
          // CREATE
          "0005 NO failed\r\n",
        };

        var presetRequest = (Action<ImapWebRequest>)delegate(ImapWebRequest req) {
          req.AllowCreateMailbox = true;
          req.DestinationUri = new Uri(new Uri(req.RequestUri.GetLeftPart(UriPartial.Path)), "./コピー先メールボックス");
        };

        Request(server, ImapWebRequestMethods.Copy, new[] {"IMAP4REV1"}, "?UNSEEN", methodResponses, presetRequest, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Failed);
        });

        StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
        StringAssert.EndsWith("UID COPY 3,4,5 \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
        StringAssert.EndsWith("CREATE \"INBOX/&MLMw1DD8UUgw4TD8MOsw3DDDMK8wuQ-\"\r\n", server.DequeueRequest());
      }
    }

    [Test]
    public void TestGetExpungeResponseStoreDeletedFlagFailure()
    {
      GetExpungeResponseStoreDeletedFlagFailure(false);
    }

    [Test]
    public void TestGetExpungeResponseStoreDeletedFlagFailureSearchresCapable()
    {
      GetExpungeResponseStoreDeletedFlagFailure(true);
    }

    private void GetExpungeResponseStoreDeletedFlagFailure(bool searchresCapable)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          searchresCapable
            ? "0003 OK done\r\n"
            : "* SEARCH 3 4 5\r\n" + 
              "0003 OK done\r\n",
          // UID STORE
          "0004 NO failed\r\n"
        };

        Request(server, ImapWebRequestMethods.Expunge, searchresCapable, "?UNSEEN", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.IsTrue(response.Result.Failed);
        });

        if (searchresCapable) {
          StringAssert.EndsWith("UID SEARCH RETURN (SAVE) UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID STORE $ +FLAGS.SILENT (\\Deleted)\r\n", server.DequeueRequest());
        }
        else {
          StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
          StringAssert.EndsWith("UID STORE 3,4,5 +FLAGS.SILENT (\\Deleted)\r\n", server.DequeueRequest());
        }
      }
    }

    [Test]
    public void TestGetExpungeResponseFailure()
    {
      const bool uidplusCapable = false;
      const bool expungeFailure = true;

      GetExpungeResponse(uidplusCapable, expungeFailure);
    }

    [Test]
    public void TestGetExpungeResponseFailureUidplusCapable()
    {
      const bool uidplusCapable = true;
      const bool expungeFailure = true;

      GetExpungeResponse(uidplusCapable, expungeFailure);
    }

    [Test]
    public void TestGetExpungeResponse()
    {
      const bool uidplusCapable = false;
      const bool expungeFailure = false;

      GetExpungeResponse(uidplusCapable, expungeFailure);
    }

    [Test]
    public void TestGetExpungeResponseUidplusCapable()
    {
      const bool uidplusCapable = true;
      const bool expungeFailure = false;

      GetExpungeResponse(uidplusCapable, expungeFailure);
    }

    private void GetExpungeResponse(bool uidplusCapable, bool expungeFailure)
    {
      using (var server = new ImapPseudoServer()) {
        var methodResponses = new[] {
          // UID SEARCH
          "* SEARCH 3 4 5\r\n" + 
          "0003 OK done\r\n",
          // UID STORE
          "0004 OK done\r\n",
          // EXPUNGE/UID EXPUNGE
          expungeFailure
          ? "0005 NO failed\r\n"
          : "* 1 EXPUNGE\r\n" +
            "* 1 EXPUNGE\r\n" +
            "* 1 EXPUNGE\r\n" +
            "0005 OK done\r\n",
        };
        var capabilities = uidplusCapable
          ? new[] {"IMAP4REV1", "UIDPLUS"}
          : new[] {"IMAP4REV1"};

        Request(server, ImapWebRequestMethods.Expunge, capabilities, "?UNSEEN", methodResponses, delegate(ImapWebRequest request, ImapWebResponse response) {
          Assert.AreEqual(request.RequestUri, response.ResponseUri);

          Assert.IsNotNull(response);
          Assert.IsNotNull(response.Result);
          Assert.AreNotEqual(response.Result.Succeeded, expungeFailure);
        });

        StringAssert.EndsWith("UID SEARCH UNSEEN\r\n", server.DequeueRequest());
        StringAssert.EndsWith("UID STORE 3,4,5 +FLAGS.SILENT (\\Deleted)\r\n", server.DequeueRequest());

        if (uidplusCapable)
          StringAssert.EndsWith("UID EXPUNGE 3,4,5\r\n", server.DequeueRequest());
        else
          StringAssert.EndsWith("EXPUNGE\r\n", server.DequeueRequest());
      }
    }
  }
}
